"""
Constants used in UConnRCMPy.
"""

one_atm_in_torr = 760.0
"""`float`: Conversion between atmosphere and Torr."""
one_atm_in_bar = 1.01325
"""`float`: Conversion between atmosphere and bar."""
one_bar_in_pa = 1.0E5
"""`float`: Conversion between bar and Pascal."""
